/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "server.h"
#include "util/memory.h"
#include "assert/assert.h"
#include "console/console.h"

server_t SERVER;

void serverInit() {
  memoryZero(&SERVER, sizeof(server_t));
  SERVER.state = SERVER_STATE_STOPPED;
}

errorret_t serverStart(uint16_t port) {
  if(SERVER.state != SERVER_STATE_STOPPED) {
    return error("Server is already running");
  }

  consolePrint("Starting server on port %d...\n", port);
  memoryZero(&SERVER, sizeof(server_t));

  // Initialize the server socket
  SERVER.serverSocket = socket(AF_INET, SOCK_STREAM, 0);
  if(SERVER.serverSocket < 0) {
    return error("Failed to create socket");
  }
  SERVER.serverAddress.sin_family = AF_INET;
  SERVER.serverAddress.sin_addr.s_addr = INADDR_ANY;
  SERVER.serverAddress.sin_port = htons(port);

  // Bind the socket to the address and port
  int32_t res = bind(
    SERVER.serverSocket,
    (struct sockaddr *)&SERVER.serverAddress,
    sizeof(SERVER.serverAddress)
  );
  if(res != 0) {
    close(SERVER.serverSocket);
    return error("Failed to bind socket");
  }

  // Set the socket to listen for incoming connections
  res = listen(SERVER.serverSocket, SERVER_MAX_CLIENTS);
  if(res != 0) {
    close(SERVER.serverSocket);
    return error("Failed to listen on socket");
  }

  // Start the server thread.
  SERVER.state = SERVER_STATE_STARTING;
  res = pthread_create(&SERVER.thread, NULL, serverThread, NULL);
  if(res != 0) {
    perror("Failed to create server thread");
    serverStop();
    return error("Failed to create server thread");
  }

  // Wait
  while(SERVER.state == SERVER_STATE_STARTING) {
    usleep(1000);
  }
  
  consolePrint("Server started.");
  return ERROR_OK;
}

void * serverThread(void *arg) {
  struct timeval timeout;
  fd_set readfds;

  SERVER.state = SERVER_STATE_RUNNING;

  while(SERVER.state == SERVER_STATE_RUNNING) {
    usleep(1000);

    FD_ZERO(&readfds);
    FD_SET(SERVER.serverSocket, &readfds);

    // Fix timeout struct
    timeout.tv_sec = 1;
    timeout.tv_usec = 0;

    // Wait for 
    int32_t activity = select(
      SERVER.serverSocket + 1,
      &readfds,
      NULL,
      NULL,
      &timeout
    );

    // Check for errors
    if(activity < 0) {
      consolePrint("Error in select");
      continue;
    }

    // Timeout
    if(activity == 0) continue;

    // Check if there is a new connection
    if(!FD_ISSET(SERVER.serverSocket, &readfds)) continue;

    // Accept new connection
    int32_t clientSocket = accept(SERVER.serverSocket, NULL, NULL);
    if(clientSocket < 0) {
      consolePrint("Failed to accept connection");
      continue;
    }

    // Initialize the client
    serverclient_t *client = NULL;
    uint8_t i = 0;
    do {
      if(SERVER.clients[i].state != SERVER_CLIENT_STATE_DISCONNECTED) {
        i++;
        continue;
      }
      client = &SERVER.clients[i];
      break;
    } while(i < SERVER_MAX_CLIENTS);

    // Can we receive this client?
    if(!client) {
      const char_t *errMsg = "ERR|Server full";
      send(clientSocket, errMsg, strlen(errMsg), 0);
      consolePrint("Server full, closing connection.");
      close(clientSocket);
      continue;
    }

    errorret_t ret = serverClientAccept(client, clientSocket);
    if(ret != ERROR_OK) {
      consolePrint("Failed to accept client connection", errorString());
      errorFlush();
      close(clientSocket);
    }
  }

  printf("Server thread exiting.\n");
  assertTrue(
    SERVER.state != SERVER_STATE_RUNNING,
    "Server thread exiting while server is running?"
  );

  SERVER.thread = 0;
  SERVER.state = SERVER_STATE_STOPPED;
  
  return NULL;
}

uint8_t serverGetClientCount() {
  uint8_t i = 0, clientCount = 0;
  do {
    if(SERVER.clients[i].state == SERVER_CLIENT_STATE_DISCONNECTED) continue;
    clientCount++;
  } while(i < SERVER_MAX_CLIENTS);
  return clientCount;
}

void serverStop() {
  if(SERVER.state == SERVER_STATE_STOPPED) return;

  assertTrue(
    SERVER.state != SERVER_STATE_STARTING,
    "Stopping a server that is starting?\n"
  );

  consolePrint("Stopping server...");
  SERVER.state = SERVER_STATE_STOPPING;

  // Wait for the server thread to finish
  int32_t maxWaits = 0;
  while(SERVER.state == SERVER_STATE_STOPPING) {
    usleep(1000);
    maxWaits++;
    if(maxWaits < 1000) continue;

    consolePrint("Server thread did not stop in time, forcing exit.");
    SERVER.state = SERVER_STATE_STOPPED;
    break;
  }

  // Kill the clients
  uint8_t i = 0;
  do {
    serverclient_t *client = &SERVER.clients[i++];
    serverClientClose(client);
  } while(i < SERVER_MAX_CLIENTS);

  // Stop the server
  if(SERVER.serverSocket >= 0) {
    close(SERVER.serverSocket);
    SERVER.serverSocket = -1;
  }

  if(SERVER.thread) {
    pthread_join(SERVER.thread, NULL);
    SERVER.thread = 0;
  }

  consolePrint("Server stopped.");
}

void serverDispose() {
  serverStop();
}