/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "serverclient.h"
#include "server.h"
#include "assert/assert.h"
#include "util/memory.h"
#include <sys/socket.h>
#include "console/console.h"

errorret_t serverClientAccept(
  serverclient_t *client,
  const int32_t socket
) {
  memoryZero(client, sizeof(serverclient_t));
  client->socket = socket;
  client->state = SERVER_CLIENT_STATE_ACCEPTING;

  // Set timeout to 8 seconds
  client->timeout.tv_sec = 8;
  client->timeout.tv_usec = 0;

  // Create a thread for the client
  if(pthread_create(&client->thread, NULL, serverClientThread, client) != 0) {
    client->state = SERVER_CLIENT_STATE_DISCONNECTED;
    return error("Failed to create client thread");
  }
}

errorret_t serverClientSend(
  serverclient_t * client,
  const uint8_t *data,
  const size_t len
) {
  assertNotNull(client, "Client is NULL");
  assertNotNull(data, "Data is NULL");
  assertTrue(len > 0, "Data length is 0");

  if(client->state != SERVER_CLIENT_STATE_ACCEPTING) {
    return error("Client is not accepting");
  }
  ssize_t sent = send(client->socket, data, len, 0);
  if(sent < 0) return error("Failed to send data");
  return ERROR_OK;
}

errorret_t serverClientSendString(
  serverclient_t * client,
  const char_t *data
) {
  return serverClientSend(client, (const uint8_t *)data, strlen(data));
}

errorret_t serverClientSendPacket(
  serverclient_t * client,
  const packet_t *packet
) {
  assertNotNull(client, "Client is NULL");
  assertNotNull(packet, "Packet is NULL");
  assertTrue(packet->type != PACKET_TYPE_INVALID, "Packet type is INVALID");
  assertTrue(packet->length > 0, "Packet length is 0");
  assertTrue(
    packet->length <= sizeof(packetdata_t),
    "Packet length is too large (1)"
  );

  size_t fullSize = sizeof(packet_t) - sizeof(packet->data) + packet->length;
  assertTrue(fullSize <= sizeof(packet_t), "Packet size is too large (2)");

  return serverClientSend(client, (const uint8_t *)packet, fullSize);
}

ssize_t serverClientReceive(
  serverclient_t * client,
  uint8_t *buffer,
  const size_t len
) {
  assertNotNull(client, "Client is NULL");
  assertNotNull(buffer, "Buffer is NULL");
  assertTrue(len > 0, "Buffer length is 0");

  if(client->state == SERVER_CLIENT_STATE_DISCONNECTED) return -1;
  if(client->state == SERVER_CLIENT_STATE_DISCONNECTING) return -1;

  return recv(client->socket, buffer, len, 0);
}

void serverClientClose(serverclient_t *client) {
  assertNotNull(client, "Client is NULL");
  if(client->state == SERVER_CLIENT_STATE_DISCONNECTED) return;

  // Mark client as disconnecting
  client->state = SERVER_CLIENT_STATE_DISCONNECTING;

  int32_t maxWaits = 0;
  while(client->state == SERVER_CLIENT_STATE_DISCONNECTING) {
    // Wait for the thread to finish
    usleep(1000);
    maxWaits++;
    if(maxWaits > 10) break;
  }

  pthread_join(client->thread, NULL);
  client->thread = 0;
  
  // Close the socket
  if(client->socket != -1) {
    close(client->socket);
    client->socket = -1;
  }

  client->state = SERVER_CLIENT_STATE_DISCONNECTED;
  consolePrint("Client %d disconnected.", client->socket);
}

void serverClientCloseOnThread(serverclient_t *client, const char_t *reason) {
  assertNotNull(client, "Client is NULL");
  assertNotNull(reason, "Reason is NULL");

  client->state = SERVER_CLIENT_STATE_DISCONNECTING;

  close(client->socket);
  client->socket = -1;

  client->state = SERVER_CLIENT_STATE_DISCONNECTED;
  client->thread = 0;

  consolePrint("Client %d disconnected: %s", client->socket, reason);
  pthread_exit(NULL);
}

void *serverClientThread(void *arg) {
  assertNotNull(arg, "Client is NULL");

  serverclient_t *client = (serverclient_t *)arg;
  char_t buffer[1024];
  ssize_t read;
  errorret_t err;
  packet_t packet;

  consolePrint("Accepting client on socket %d.", client->socket);

  // Set socket timeout
  if(setsockopt(
    client->socket,
    SOL_SOCKET,
    SO_RCVTIMEO,
    &client->timeout,
    sizeof(client->timeout)
  ) < 0) {
    serverClientCloseOnThread(client, "Failed to set socket timeout");
    return NULL;
  }

  // First message from the client should always be "DUSK|VERSION" to match
  // the server version.
  {
    const char_t *expecting = "DUSK|"DUSK_VERSION;
    read = serverClientReceive(client, buffer, sizeof(buffer));
    if(read <= 0) {
      packetDisconnectCreate(&packet, PACKET_DISCONNECT_REASON_INVALID_VERSION);
      err = serverClientSendPacket(client, &packet);
      serverClientCloseOnThread(client, "Failed to receive version");
      return NULL;
    }

    buffer[read] = '\0'; // Null-terminate the string
    if(strncmp(buffer, expecting, strlen(expecting)) != 0) {
      packetDisconnectCreate(&packet, PACKET_DISCONNECT_REASON_INVALID_VERSION);
      err = serverClientSendPacket(client, &packet);
      serverClientCloseOnThread(client, "Invalid version");
      return NULL;
    }
  }

  // Send DUSK back!
  packetWelcomeCreate(&packet);
  err = serverClientSendPacket(client, &packet);
  if(err != ERROR_OK) {
    serverClientCloseOnThread(client, "Failed to send welcome message");
    return NULL;
  }

  client->state = SERVER_CLIENT_STATE_CONNECTED;
  while(client->state == SERVER_CLIENT_STATE_CONNECTED) {
    read = serverClientReceive(client, buffer, sizeof(buffer));

    if(read <= 0) {
      serverClientCloseOnThread(client, "Failed to receive data");
      break;
    }

    buffer[read] = '\0'; // Null-terminate the string
    consolePrint("Received: %s", buffer);

    
    if(SERVER.state != SERVER_STATE_RUNNING) {
      serverClientCloseOnThread(client, "Server is shutting down");
      break;
    }
  }

  serverClientSendString(client, "Server is shutting down");
  client->state = SERVER_CLIENT_STATE_DISCONNECTED;
  return NULL;
}