/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#pragma once
#include "error/error.h"
#include "server/packet/packet.h"

typedef enum {
  SERVER_CLIENT_STATE_DISCONNECTED,
  SERVER_CLIENT_STATE_ACCEPTING,
  SERVER_CLIENT_STATE_CONNECTED,
  SERVER_CLIENT_STATE_DISCONNECTING,
} serverclientstate_t;

typedef struct {
  int32_t socket;
  serverclientstate_t state;
  pthread_t thread;
  struct timeval timeout;
} serverclient_t;

/**
 * Accepts an incoming connection from a server client.
 * 
 * @param client Pointer to the server client structure.
 * @param socket The socket file descriptor for the client.
 * @return Error code indicating success or failure.
 */
errorret_t serverClientAccept(
  serverclient_t *client,
  const int32_t socket
);

/**
 * Sends data to a server client.
 * 
 * @param client Pointer to the server client structure.
 * @param data Pointer to the data to send.
 * @param len Length of the data to send.
 * @return Error code indicating success or failure.
 */
errorret_t serverClientSend(
  serverclient_t * client,
  const uint8_t *data,
  const size_t len
);

/**
 * Sends a string to a server client.
 * 
 * @param client Pointer to the server client structure.
 * @param data Pointer to the string to send.
 * @return Error code indicating success or failure.
 */
errorret_t serverClientSendString(
  serverclient_t * client,
  const char_t *data
);

/**
 * Sends a packet to a server client.
 * 
 * @param client Pointer to the server client structure.
 * @param packet Pointer to the packet to send.
 * @return Error code indicating success or failure.
 */
errorret_t serverClientSendPacket(
  serverclient_t * client,
  const packet_t *packet
);

/**
 * Receives data from a server client.
 * 
 * @param client Pointer to the server client structure.
 * @param buffer Pointer to the buffer to store received data.
 * @param len Max length of the buffer.
 * @return Number of bytes received. 0 or less indicates an error.
 */
ssize_t serverClientReceive(
  serverclient_t * client,
  uint8_t *buffer,
  const size_t len
);

/**
 * Closes the connection to a server client. Waits for the thread to finish.
 * 
 * @param client Pointer to the server client structure.
 */
void serverClientClose(serverclient_t *client);

/**
 * Closes the connection to a server client without waiting for the thread to
 * finish.
 * 
 * @param client Pointer to the server client structure.
 * @param reason Reason for closing the connection.
 */
void serverClientCloseOnThread(
  serverclient_t *client,
  const char_t *reason
);

/**
 * Thread function for handling a server client.
 * 
 * @param arg Pointer to the server client structure.
 * @return NULL.
 */
void *serverClientThread(void *arg);